#!/usr/local/bin/python3

# Booze design raw rgb 160x200 video to koala stream converter
# Based on c64image.py by Philip Heron, but stripped unwanted stuff, plus some
# optimizations.
# Use pypy3 for speed!

# c64image.py
#
# A small python program to convert an image to Commodore 64 multi-colour
# bitmap data. Output can be an executable PRG or a header file.
#
# - Philip Heron <phil@sanslogic.co.uk>
#
# --
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# --

import sys
import math
from PIL import Image

palette = (
    (0x00, 0x00, 0x00),
    (0xFF, 0xFF, 0xFF),
    (0x8C, 0x3E, 0x34),
    (0x7A, 0xBF, 0xC7),
    (0x8D, 0x47, 0xB3),
    (0x68, 0xA9, 0x41),
    (0x3E, 0x31, 0xA2),
    (0xD0, 0xDC, 0x71),
    (0x90, 0x5F, 0x25),
    (0x57, 0x42, 0x00),
    (0xBB, 0x77, 0x6D),
    (0x54, 0x54, 0x54),
    (0x80, 0x80, 0x80),
    (0xAC, 0xEA, 0x88),
    (0x7C, 0x70, 0xDA),
    (0xAB, 0xAB, 0xAB),
)

# Convert an (r, g, b) value to the nearest C64 palette entry
def rgb2pal(colour):
    
    s = -1
    c = -1
    
    for x in range(0, len(palette)):
        d = (
            (colour[0] - palette[x][0]) ** 2 +
            (colour[1] - palette[x][1]) ** 2 +
            (colour[2] - palette[x][2]) ** 2
        )
        if s == -1 or d < s:
            s = d
            c = x
    
    return c

# Find the nearest C64 palette entry from a list of palette codes
def pal2pal(colour, colours):
    
    s = -1
    c = -1
    
    colour = palette[colour]
    
    for x in range(0, len(colours)):
        d = (
            (colour[0] - palette[colours[x]][0]) ** 2 +
            (colour[1] - palette[colours[x]][1]) ** 2 +
            (colour[2] - palette[colours[x]][2]) ** 2
        )

        if s == -1 or d < s:
            s = d
            c = colours[x]
    
    return c

# Write bytes to a string
def write_bytes(name, data, cformat):
    
    if cformat == 'S':
        s = "%s\n" % name
        
        for line in (data[x:x + 32] for x in range(0, len(data), 32)):
            s += "\t.byte " + ",".join(("$%02X" % x for x in line)) + "\n"
    
    elif cformat == 'H':
        s = "unsigned char %s[0x%d] = [\n" % (name, len(data))
        
        for line in (data[x:x + 32] for x in range(0, len(data), 32)):
            s += "\t" + ",".join(("0x%02X" % x for x in line)) + ",\n"
        
        s += "];\n";
    
    return s

canvas = [[[[0] * 4 for row in range(8)] for ccol in range(40)] for crow in range(25)]
   
while True:

    # Open and convert the image to 160x200 RGB frame
    indata = sys.stdin.buffer.read(160*200*3)
    im = Image.frombytes('RGB', (160, 200), indata)

    # Colour usage counters, palette order
    ccounter = [0] * 16
    
    # Render the PIL image to the canvas
    for cy in range(0, 25):
        for cx in range(0, 40):
            
            ox = cx * 4
            oy = cy * 8
            
            for y in range(0, 8):
                for x in range(0, 4):
                    
                    p = rgb2pal(im.getpixel((ox + x, oy + y)))
                    canvas[cy][cx][y][x] = p
                    ccounter[p] += 1
    
    #Select the background based on the most used colour
    background = ccounter.index(max(ccounter))
    
    # Initialise the bitmap buffer
    bitmap_bytes = []
    screen_bytes = []
    colour_bytes = []
    
    last_colours = False
    
    for cy in range(0, 25):
        for cx in range(0, 40):
            
            # Make a list of all the colours used in this block
            # Background is always used, even if not referenced
            colours = [background]
            ccounter = [0]
            
            for row in canvas[cy][cx]:
                for colour in row:
                    
                    if not colour in colours:
                        colours.append(colour)
                        ccounter.append(0)
                    
                    ccounter[colours.index(colour)] += 1
            
            # Sort the colours, background first then most used to least
            colours = [background] + list(x[1] for x in sorted(zip(ccounter[1:], colours[1:]), reverse = True))
            
            if len(colours) > 4:
                
                # Crop colour list to the background and 3 most used values
                colours = colours[:4]
                
                # Replace removed colours in the block with the nearest valid match
                for y in range(8):
                    for x in range(4):
                        
                        p = canvas[cy][cx][y][x]
                        
                        if not p in colours:
                            canvas[cy][cx][y][x] = pal2pal(p, colours)
            
            # If all the colours of this block are present in the previous block,
            # use that same order. This may help with RLE compression.
            if last_colours != False and set(colours).issubset(set(last_colours)):
                colours = last_colours
            
            elif len(colours) != 4:
                colours += [0] * (4 - len(colours))
            
            last_colours = colours
            
            # Output this block
            for row in canvas[cy][cx]:
                bitmap_bytes.append(colours.index(row[0]) << 6 | colours.index(row[1]) << 4 | colours.index(row[2]) << 2 | colours.index(row[3]))
            
            screen_bytes.append(colours[1] << 4 | colours[2])
            colour_bytes.append(colours[3])
    
    # Write the results
    sys.stdout.buffer.write(bytes(bitmap_bytes) + bytes(screen_bytes) + bytes(colour_bytes) + bytes([background]))
